# -*- coding: utf-8 -*-


class FuzzyMode():

    """docstring for FuzzyMode"""

    def __init__(self, closeness_obj):
        self.ARRAY_CMP_FIELDS = closeness_obj.ARRAY_CMP_FIELDS
        self.ARRAY_DICT_CMP_FIELDS = closeness_obj.ARRAY_DICT_CMP_FIELDS
        self.STRING_CMP_FIELDS = closeness_obj.STRING_CMP_FIELDS
        self.NUM_CMP_FIELDS = closeness_obj.NUM_CMP_FIELDS
        self.cmp_object = closeness_obj.cmp_object
        self.unit_weight = closeness_obj.unit_weight

    def get_pipeline_stage3_group(self):
        query = {'$group': {
            '_id': '$_id',
            'rank': {'$sum': {'$add': []}},
            'weights': {'$push': {}},
        }}
        self.set_nested_field_weightages(query)
        self.set_string_field_weightages(query)
        self.set_int_field_weightages(query)

        return query

    def set_nested_field_weightages(self, query):
        for var in self.ARRAY_CMP_FIELDS + self.ARRAY_DICT_CMP_FIELDS:
            self_value = self.cmp_object.get(var['field'])
            if self_value:
                field_weight = self.get_formula(var, self_value)
                query['$group']['rank']['$sum']['$add'].append(
                    field_weight
                )
                self.mark_weights(query, var['field'], field_weight)

        return query

    def set_string_field_weightages(self, query):
        for var in self.STRING_CMP_FIELDS:
            self_value = self.cmp_object.get(var['field'])
            if self_value:
                # equation = {'$cond': [conition, true value, false value]}
                equation = {
                    '$cond': [{}, var.get('weight', 1) * self.unit_weight, 0]
                }

                # compare string
                equation['$cond'][0]['$eq'] = [
                    {'$strcasecmp': ["$" + var['field'], self_value]}, 0
                ]

                query['$group']['rank']['$sum']['$add'].append(equation)
                self.mark_weights(query, var['field'], equation)

        return query

    def set_int_field_weightages(self, query):
        for var in self.NUM_CMP_FIELDS:
            self_value = self.cmp_object.get(var['field'])
            if self_value:
                # equation = {'$cond': [conition, true value, false value]}
                equation = {
                    '$cond': [{}, var.get('weight', 1) * self.unit_weight, 0]
                }
                # check number is between given range
                equation['$cond'][0]['$and'] = [{
                    '$gte': ['$' + var['field'], self_value + var['from']]
                }, {
                    '$lte': ['$' + var['field'], self_value + var['to']]
                }]

                query['$group']['rank']['$sum']['$add'].append(equation)
                self.mark_weights(query, var['field'], equation)

        return query

    def set_field_weight(self, var, self_value):
        """
            weight calculation formula :

            (( 2 * ( n(A ⋂ B) / n(A) ) + n(A ⋂ B) / n(B) ) / 3 ) * weight


        """

        # n(A)
        nA = len(self_value)

        if nA == 0:
            return {}

        # n(B)     adding 0.001 to avoid divsible by 0 error
        nB = {'$add': [
            {'$size': '$' + var['field']},
            0.001
        ]}

        # n(A ⋂ B)
        nAandB = {'$size': {
            '$setIntersection': [
                self_value,
                '$' + var['field']
            ],
        }}

        # ( 2 * ( n(A ⋂ B) / n(A) )

        Two_nAandBby_nA = {
            '$multiply': [{
                '$divide': [nAandB, nA],
            }, 2],
        }

        # n(A ⋂ B) / n(B)

        nAandBby_nA = {
            '$divide': [nAandB, nB]
        }

        # ( 2 * ( n(A U B) / n(A) ) + n(A ⋂ B) / n(B) ) / 3
        Two_nAandBby_nA_Plus_nAandBby_nA_By_three = {
            '$divide': [
                {'$add': [Two_nAandBby_nA, nAandBby_nA]},
                3
            ],
        }

        return Two_nAandBby_nA_Plus_nAandBby_nA_By_three

    def get_formula(self, var, self_value):
        Two_nAandBby_nA_Plus_nAandBby_nA_By_three = self.set_field_weight(
            var,
            self_value
        )

        # (( 2 * ( n(A ⋂ B) / n(A) ) + n(A ⋂ B) / n(B) ) / 3 ) * weight
        Two_nAandBby_nA_Plus_nAandBby_nA_By_three_Mul_weight = {
            '$multiply': [
                Two_nAandBby_nA_Plus_nAandBby_nA_By_three,
                var.get('weight', 1) * self.unit_weight
            ],
        }

        return Two_nAandBby_nA_Plus_nAandBby_nA_By_three_Mul_weight

    def mark_weights(self, query, field, field_weight):
        query['$group']['weights']['$push'][field] = {
            '$add': [
                field_weight,
                0
            ]
        }
        return query



# -*- coding: utf-8 -*-

from collections import OrderedDict
from fuzzy_mode import FuzzyMode
from simple_mode import SimpleMode


class ClosenessAggregation():

    (SIMPLE, FUZZY) = (1, 2)

    def __init__(
            self, cmp_object, query, out_fields, limit, page=1, **cmp_fields):
        self.cmp_object = cmp_object
        self.page = page
        self.limit = limit
        self.query = query
        self.pipeline = []
        self.set_cmp_feilds(cmp_fields)
        self.OUT_PUT_FIELDS = out_fields

    def set_cmp_feilds(self, cmp_fields):
        weight = 0
        count = 0
        for option, value in cmp_fields.iteritems():
            setattr(self, option, value)
            for field in value:
                weight += field.get('weight', 1)
                count += 1

        self.unit_weight = 100 / weight

    def get_aggregation_pipeline(self, mode=None):
        self.set_mode(mode)

        self.generate_pipeline()
        return self.pipeline

    def set_mode(self, mode):
        if mode == self.SIMPLE:
            self.mode = SimpleMode(self)
        else:
            self.mode = FuzzyMode(self)

    def generate_pipeline(self):
        self.set_pipeline_stage1_query()
        self.set_pipeline_stage2_project()
        self.set_pipeline_stage3_group()
        self.set_pipeline_stage4_sort()
        self.set_pipeline_limit_and_skip_stage()

    def set_pipeline_stage1_query(self):
        if self.query:
            self.pipeline.append(self.query)

    def set_pipeline_stage2_project(self):
        self.format_self_user_emb_docs()
        query = {'$project': {}}
        for field in self.cmp_object:
            query['$project'][field] = 1

        for array_dict in self.ARRAY_DICT_CMP_FIELDS:
            field = array_dict['field']
            uniq_id = array_dict['unique']
            query['$project'][field] = {
                '$map': {
                    'input': "$" + field,
                    'as': "temp",
                    'in': "$$temp." + uniq_id
                }
            }

        self.pipeline.append(query)

    def set_pipeline_stage3_group(self):
        query = self.mode.get_pipeline_stage3_group()
        query = self.set_output_fields(query)
        self.pipeline.append(query)

    def set_output_fields(self, query):
        for field in self.OUT_PUT_FIELDS:
            query['$group'][field] = {'$first': '$' + field}
        return query

    def set_pipeline_stage4_sort(self):
        sort_dict = OrderedDict()
        sort_dict['rank'] = -1
        sort_dict['details.distance'] = 1
        sort_dict['_id'] = -1

        query = {'$sort': sort_dict}
        self.pipeline.append(query)

    def set_pipeline_limit_and_skip_stage(self):
        self.pipeline.append(
            {'$skip': (self.page - 1) * self.limit},
        )
        self.pipeline.append(
            {'$limit': self.limit},
        )

    def format_self_user_emb_docs(self):
        for array_dict in self.ARRAY_DICT_CMP_FIELDS:
            uniq_id = array_dict['unique']
            field = array_dict['field']
            ids = []
            for doc in self.cmp_object[field]:
                ids.append(doc[uniq_id])
            self.cmp_object[field] = ids



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'DESCRIPTION.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='closeness',

    version='1.5.0',

    description='Mongodb document similarity',
    long_description=long_description,

    url='https://github.com/mthahakp/mongodb-doc-closeness',

    author='Thaha',
    author_email='tkannippoyil@suyati.com',

    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],

    keywords='mongodb setuptools development relationship query',

    packages=['closeness'],

)
